@@ -14,6 +14,9 @@ from scipy.fftpack import next_fast_len
 def exists(val):
     return val is not None
 
+def identity(t, *args, **kwargs):
+    return t
+
 def default(val, d):
     return val if exists(val) else d
 
@@ -189,8 +192,8 @@ class MultiHeadedEMA(nn.Module):
         super().__init__()
         self.bidirectional = bidirectional
 
-        self.expansion = nn.Parameter(torch.randn(heads, dim))
-        self.reduction = nn.Parameter(torch.randn(heads, dim))
+        self.expansion = nn.Parameter(torch.randn(heads * (2 if bidirectional else 1), dim))
+        self.reduction = nn.Parameter(torch.randn(heads * (2 if bidirectional else 1), dim))
 
         # learned alpha and dampening factors
 
@@ -208,6 +211,10 @@ class MultiHeadedEMA(nn.Module):
 
         x = einsum('... d, h d -> ... h d', x, self.expansion)
 
+        if self.bidirectional:
+            x, x_reversed = x.chunk(2, dim = -2)
+            x_reversed = torch.flip(x_reversed, dims = (1,))
+
         # weights derived from alphas (learned exponential smoothing decay rate)
 
         def apply_learned_ema_with_damping(x, alphas, dampen_factors):
@@ -224,9 +231,9 @@ class MultiHeadedEMA(nn.Module):
         x = apply_learned_ema_with_damping(x, self.alphas, self.dampen_factors)
 
         if self.bidirectional:
-            x = torch.flip(x, dims = (1,))
-            x = apply_learned_ema_with_damping(x, self.reverse_alphas, self.reverse_dampen_factors)
-            x = torch.flip(x, dims = (1,))
+            x_reversed = apply_learned_ema_with_damping(x_reversed, self.reverse_alphas, self.reverse_dampen_factors)
+            x_reversed = torch.flip(x_reversed, dims = (1,))
+            x = torch.cat((x, x_reversed), dim = -2)
 
         # combine heads and out
 
@@ -280,7 +287,9 @@ class MegaLayer(nn.Module):
         self.Uh = nn.Parameter(torch.randn(attn_dim_value, dim))
         self.bh = nn.Parameter(torch.randn(dim))
 
-    def forward(self, x):
+    def forward(self, x, residual = None):
+        residual = default(residual, x)
+
         ema_output = self.multi_headed_ema(x)
         attn_output = self.single_headed_attn(ema_output, x)
 
@@ -315,30 +324,46 @@ class Mega(nn.Module):
         num_tokens,
         depth,
         ff_mult = 2,
+        pre_norm = False,
         **kwargs
     ):
         super().__init__()
         self.token_emb = nn.Embedding(num_tokens, dim)
+        self.pre_norm = pre_norm
 
         self.layers = nn.ModuleList([])
         for _ in range(depth):
             self.layers.append(nn.ModuleList([
-                MegaLayer(**kwargs),
+                MegaLayer(dim = dim, **kwargs),
                 nn.LayerNorm(dim),
                 FeedForward(dim = dim, ff_mult = ff_mult),
                 nn.LayerNorm(dim)
             ]))
 
-        self.to_logits = nn.Linear(dim, num_tokens)
+        self.to_logits = nn.Sequential(
+            nn.LayerNorm(dim) if pre_norm else nn.Identity(),
+            nn.Linear(dim, num_tokens)
+        )
 
     def forward(self, x):
+        pre_norm = self.pre_norm
+        post_norm = not self.pre_norm
+
         x = self.token_emb(x)
 
-        for mega_layer, post_mega_norm, ff, post_ff_norm in self.layers:
-            x = mega_layer(x)
-            x = post_mega_norm(x)
+        for mega_layer, mega_norm, ff, ff_norm in self.layers:
+            mega_maybe_prenorm = mega_norm if pre_norm else identity
+            ff_maybe_prenorm = ff_norm if pre_norm else identity
 
-            x = ff(x) + x
-            x = post_ff_norm(x)
+            mega_maybe_postnorm = mega_norm if post_norm else identity
+            ff_maybe_postnorm = ff_norm if post_norm else identity
+
+            x = mega_layer(mega_maybe_prenorm(x), x)
+
+            x = mega_maybe_postnorm(x)
+
+            x = ff(ff_maybe_prenorm(x)) + x
+
+            x = ff_maybe_postnorm(x)
 
         return self.to_logits(x)
